import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:bloc_test/bloc_test.dart';
import 'package:test/test.dart';

import 'data/dummy_bloc.dart';
import 'data/dummy_event.dart';
import 'data/dummy_state.dart';

void main() {
  DummyBloc bloc;

  setUp(() {
    BlocSupervisor.delegate = _LoggingBlocDelegate();
    bloc = DummyBloc();
  });

  tearDown(() async {
    await bloc.close();
  });

  group("StateGraphBlocTest", () {
    test('should emmit proper states', () async {
      // When
      bloc.add(LoadEvent());

      // Then
      await emitsExactly(
          bloc,
          [
            StateInitialising(),
            StateLoading(),
            StateReady(0),
          ],
          skip: 0,
          duration: Duration(seconds: 1));
    });

    test('should emmit proper transformed data', () async {
      // Given
      final loading = bloc.isLoading();

      // When
      bloc.add(LoadEvent());
      bloc.add(LoadingCompletedEvent());

      // Then
      await expectLater(
        loading,
        emitsInOrder([
          false, // StateInitialising
          true, // StateLoading
          false, // StaeReady
        ]),
      );
    });

    test('should use global events', () async {
      // When
      bloc.add(LoadEvent());

      // Then
      await expectLater(
          bloc,
          emitsInOrder([
            StateInitialising(),
            StateLoading(),
            StateReady(0),
          ]));

      // And
      bloc.add(IncrementEvent(1));
      bloc.add(RestartEvent());

      // Then
      await expectLater(
          bloc,
          emitsInOrder([
            StateReady(0),
            StateReady(1),
            StateReady(0),
          ]));
    });

    test("should launch future", () async {
      // Given
      final future = Future.value("a");

      // When
      String receivedData;
      await bloc.launchFuture(future, onData: (data) {
        receivedData = data;
      });

      // Then
      expect(receivedData, "a");
    });

    test("should close future launched after bloc was closed", () async {
      // Given
      final future = Future.value("a");

      // When
      bloc.close();
      String receivedData;
      final subscription = bloc.launchFuture(future, onData: (data) {
        receivedData = data;
      });
      await subscription.asFuture();

      // Then
      expect(receivedData, null);
    });

    test("should launch stream subscription", () async {
      // Given
      final stream = Stream.value("a");

      // When
      String receivedData;
      final subscription = bloc.launchStream(stream, onData: (data) {
        receivedData = data;
      });
      await subscription.asFuture();

      // Then
      expect(receivedData, "a");
    });

    test("should dispose stream subscription after bloc was closed", () async {
      // Given
      final stream = Stream.value("a");

      // When
      bloc.close();
      String receivedData;
      final subscription = bloc.launchStream(stream, onData: (data) {
        receivedData = data;
      });
      await subscription.asFuture(null);

      // Then
      expect(receivedData, null);
    });
  });
}

class _LoggingBlocDelegate extends BlocDelegate {
  @override
  void onError(Bloc bloc, Object error, StackTrace stackTrace) {
    super.onError(bloc, error, stackTrace);
    print("onError: $error");
  }

  @override
  void onEvent(Bloc bloc, Object event) {
    super.onEvent(bloc, event);
    print("onEvent: $event");
  }

  @override
  void onTransition(Bloc bloc, Transition transition) {
    super.onTransition(bloc, transition);
    print("onTransition: $transition");
  }
}
